@@ -1,6 +1,7 @@
 from typing import List, Optional
 import logging
 from pathlib import Path
+from datetime import datetime
 
 from .utils import safezip, dataclass, SafeDict, listgen
 from .interp_common import assert_type, exclude_fields, call_pql_func, is_global_scope
@@ -56,10 +57,12 @@ def resolve(state: State, col_def: ast.ColumnDef):
     assert not query
 
     if isinstance(coltype, objects.SelectedColumnInstance):
-        return T.t_relation[coltype.type](rel={'table': coltype.parent.type, 'column': coltype.name, 'key': False}).replace(_nullable=coltype.type._nullable)
+        x = T.t_relation[coltype.type](rel={'table': coltype.parent.type, 'column': coltype.name, 'key': False})
+        return x.replace(_nullable=coltype.type._nullable)  # inherit is_nullable (TODO: use sumtypes?)
 
     elif coltype <= T.table:
-        return T.t_relation[T.t_id.as_nullable()](rel={'table': coltype, 'column': 'id', 'key': True}).replace(_nullable=coltype._nullable)
+        x = T.t_relation[T.t_id.as_nullable()](rel={'table': coltype, 'column': 'id', 'key': True})
+        return x.replace(_nullable=coltype._nullable)     # inherit is_nullable (TODO: use sumtypes?)
 
     return coltype(default=col_def.default)
 
@@ -201,11 +204,11 @@ def _copy_rows(state: State, target_name: ast.Name, source: objects.TableInstanc
 def _execute(state: State, insert_rows: ast.InsertRows):
     if not isinstance(insert_rows.name, ast.Name):
         # TODO support Attr
-        raise Signal.make(T.SyntaxError, state, insert_rows, "L-value must be table name")
+        raise Signal.make(T.SyntaxError, insert_rows, "L-value must be table name")
 
     rval = evaluate(state, insert_rows.value)
 
-    assert_type(rval.type, T.collection, state, insert_rows, '+=')
+    assert_type(rval.type, T.table, state, insert_rows, '+=')
 
     return _copy_rows(state, insert_rows.name, rval)
 
@@ -268,7 +271,7 @@ def _execute(state: State, i: ast.If):
 
 @dy
 def _execute(state: State, w: ast.While):
-    while(cast_to_python(state, w.cond)):
+    while cast_to_python(state, w.cond):
         execute(state, w.do)
 
 @dy
@@ -388,7 +391,6 @@ def simplify(state: State, x):
 
 # @dy
 # def simplify(state: State, node: ast.Ast):
-#     # TODO implement automatically with prerequisites
 #     # return _simplify_ast(state, node)
 #     return node
 
@@ -408,6 +410,7 @@ def simplify(state: State, x):
 #             return if_.else_
 #     return if_
 
+
 # TODO Optimize these, right now failure to evaluate will lose all work
 @dy
 def simplify(state: State, obj: ast.Or):
@@ -496,7 +499,7 @@ def eval_func_call(state, func, args):
     for i, (p, a) in enumerate(matched_args):
         a = evaluate(state, a)
         # TODO cast?
-        if p.type and not kernel_type(a.type) <= p.type:
+        if p.type and not a.type <= p.type:
             raise Signal.make(T.TypeError, func, f"Argument #{i} of '{func.name}' is of type '{a.type}', expected '{p.type}'")
         args[p.name] = a
 
@@ -506,43 +509,42 @@ def eval_func_call(state, func, args):
         # TODO Ensure correct types
         args = list(args.values())
         return func.func(state, *args)
-    else:
-        # TODO make tests to ensure caching was successful
-        if settings.cache:
-            params = {name: ast.Parameter(name, value.type) for name, value in args.items()}
-            sig = (func.name,) + tuple(a.type for a in args.values())
 
-            try:
-                with state.use_scope(params):
-                    if sig in state._cache:
-                        expr = state._cache[sig]
-                    else:
-                        logging.info(f"Compiling.. {func}")
-                        expr = _call_expr(state.reduce_access(state.AccessLevels.COMPILE), func.expr)
-                        logging.info("Compiled successfully")
-                        if isinstance(expr, objects.Instance):
-                            # XXX a little ugly
-                            qb = sql.QueryBuilder(state.db.target, True)
-                            x = expr.code.compile(qb)
-                            x = x.optimize()
-                            expr = expr.replace(code=x)
-                        state._cache[sig] = expr
+    # TODO make tests to ensure caching was successful
+    expr = func.expr
+    if settings.cache:
+        params = {name: ast.Parameter(name, value.type) for name, value in args.items()}
+        sig = (func.name,) + tuple(a.type for a in args.values())
 
-                expr = ast.ResolveParameters(expr, args)
+        try:
+            with state.use_scope(params):
+                if sig in state._cache:
+                    compiled_expr = state._cache[sig]
+                else:
+                    logging.info(f"Compiling.. {func}")
+                    compiled_expr = _call_expr(state.reduce_access(state.AccessLevels.COMPILE), func.expr)
+                    logging.info("Compiled successfully")
+                    if isinstance(compiled_expr, objects.Instance):
+                        # XXX a little ugly
+                        qb = sql.QueryBuilder(state.db.target, True)
+                        x = compiled_expr.code.compile(qb)
+                        x = x.optimize()
+                        compiled_expr = compiled_expr.replace(code=x)
+                    state._cache[sig] = compiled_expr
 
-            except exc.InsufficientAccessLevel:
-                # Don't cache
-                expr = func.expr
-        else:
-            expr = func.expr
+            expr = ast.ResolveParameters(compiled_expr, args)
 
-        with state.use_scope(args):
-            res = _call_expr(state, expr)
+        except exc.InsufficientAccessLevel:
+            # Don't cache
+            pass
 
-        if isinstance(res, ast.ResolveParameters):  # XXX A bit of a hack
-            raise exc.InsufficientAccessLevel()
+    with state.use_scope(args):
+        res = _call_expr(state, expr)
 
-        return res
+    if isinstance(res, ast.ResolveParameters):  # XXX A bit of a hack
+        raise exc.InsufficientAccessLevel()
+
+    return res
 
 
 def _call_expr(state, expr):
@@ -585,7 +587,7 @@ def apply_database_rw(state: State, o: ast.One):
     slice_ast = ast.Slice(obj, ast.Range(None, ast.Const(T.int, 2))).set_text_ref(o.text_ref)
     table = evaluate(state, slice_ast)
 
-    assert (table.type <= T.collection), table
+    assert (table.type <= T.table), table
     rows = localize(state, table) # Must be 1 row
     if len(rows) == 0:
         if not o.nullable:
@@ -614,7 +616,7 @@ def apply_database_rw(state: State, d: ast.Delete):
     cond_table = ast.Selection(d.table, d.conds).set_text_ref(d.text_ref)
     table = evaluate(state, cond_table)
 
-    if not (table.type <= T.table):
+    if not table.type <= T.table:
         raise Signal.make(T.TypeError, d.table, f"Expected a table. Got: {table.type}")
 
     if not 'name' in table.type.options:
@@ -639,7 +641,7 @@ def apply_database_rw(state: State, u: ast.Update):
     # TODO Optimize: Update on condition, not id, when possible
     table = evaluate(state, u.table)
 
-    if not (table.type <= T.table):
+    if not table.type <= T.table:
         raise Signal.make(T.TypeError, u.table, f"Expected a table. Got: {table.type}")
 
     if not 'name' in table.type.options:
@@ -720,12 +722,14 @@ def _destructure_param_match(state, ast_node, param_match):
             v = v.primary_key()
         v = localize(state, v)
 
-        if (k.type <= T.struct):
+        if k.type <= T.struct:
             names = [name for name, t in flatten_type(k.orig, [k.name])]
             if not isinstance(v, list):
-                raise Signal.make(T.TypeError, ast_node, f"Parameter {k.name} received a bad value: {v} (expecting a struct or a list)")
+                msg = f"Parameter {k.name} received a bad value: {v} (expecting a struct or a list)"
+                raise Signal.make(T.TypeError, ast_node, msg)
             if len(v) != len(names):
-                raise Signal.make(T.TypeError, ast_node, f"Parameter {k.name} received a bad value (size of {len(names)})")
+                msg = f"Parameter {k.name} received a bad value (size of {len(names)})"
+                raise Signal.make(T.TypeError, ast_node, msg)
             yield from safezip(names, v)
         else:
             yield k.name, v
@@ -810,12 +814,14 @@ class TableConstructor(objects.Function):
 
     @classmethod
     def make(cls, table):
-        return cls([objects.Param(name, p, p.options.get('default'), orig=p).set_text_ref(getattr(name, 'text_ref', None))
-                    for name, p in table_params(table)])
+        return cls([
+            objects.Param(name, p, p.options.get('default'), orig=p).set_text_ref(getattr(name, 'text_ref', None))
+            for name, p in table_params(table)
+        ])
 
 
 def add_as_subquery(state: State, inst: objects.Instance):
-    code_cls = sql.TableName if (inst.type <= T.collection) else sql.Name
+    code_cls = sql.TableName if (inst.type <= T.table) else sql.Name
     name = state.unique_name(inst)
     return inst.replace(code=code_cls(inst.code.type, name), subqueries=inst.subqueries.update({name: inst.code}))
 
@@ -944,7 +950,8 @@ def new_table_from_expr(state, name, expr, const, temporary):
         return objects.TableInstance.make(sql.null, expr.type, [])
 
     if 'id' in elems and not const:
-        raise Signal.make(T.NameError, None, "Field 'id' already exists. Rename it, or use 'const table' to copy it as-is.")
+        msg = "Field 'id' already exists. Rename it, or use 'const table' to copy it as-is."
+        raise Signal.make(T.NameError, None, msg)
 
     table = T.table(dict(elems), name=Id(name), pk=[] if const else [['id']], temporary=temporary)
 
@@ -972,9 +979,9 @@ def cast_to_python(state, obj: ast.Ast):
 @dy
 def cast_to_python(state, obj: objects.AbsInstance):
     # if state.access_level <= state.AccessLevels.QUERY:
-    if obj.type <= T.vectorized:
+    if obj.type <= T.projected | T.aggregated:
         raise exc.InsufficientAccessLevel(state.access_level)
-        # raise Signal.make(T.CastError, None, f"Internal error. Cannot cast vectorized (i.e. projected) obj: {obj}")
+        # raise Signal.make(T.CastError, None, f"Internal error. Cannot cast projected obj: {obj}")
     res = localize(state, obj)
     if obj.type == T.float:
         res = float(res)
@@ -983,7 +990,7 @@ def cast_to_python(state, obj: objects.AbsInstance):
     elif obj.type == T.bool:
         assert res in (0, 1), res
         res = bool(res)
-    assert isinstance(res, (int, str, float, dict, list, type(None))), (res, type(res))
+    assert isinstance(res, (int, str, float, dict, list, type(None), datetime)), (res, type(res))
     return res
 
 
