@@ -1,27 +1,25 @@
 import inspect
-from preql.compiler import cast_to_instance
 import re
 import os
 from datetime import datetime
 import csv
-import inspect
 import itertools
+
+import rich.progress
 import runtype
-import json
 
 from .utils import safezip, listgen, re_split
 from .exceptions import Signal, ExitInterp
-
 from . import pql_objects as objects
 from . import pql_ast as ast
 from . import sql
-
 from .interp_common import State, new_value_instance, assert_type
 from .evaluate import evaluate, cast_to_python, db_query, TableConstructor, new_table_from_expr, new_table_from_rows
-from .pql_types import T, Type, union_types, Id, common_type
+from .pql_types import T, Type, Id
 from .types_impl import join_names
 from .casts import cast
-from .docstring.autodoc import autodoc
+from .docstring.autodoc import autodoc, AutoDocError
+from .compiler import cast_to_instance
 
 def new_str(x):
     return new_value_instance(str(x), T.string)
@@ -84,11 +82,16 @@ def pql_inspect_sql(state: State, obj: T.object):
     return objects.ValueInstance.make(sql.make_value(s), T.text, [], s)
 
 
-def pql_SQL(state: State, result_type: T.union[T.collection, T.type], sql_code: T.string):
+def pql_SQL(_state: State, result_type: T.union[T.table, T.type], sql_code: T.string):
     """Create an object with the given SQL evaluation code, and given result type.
 
     The object will only be evaluated when required by the program flow.
 
+    Using $var_name in the code will embed it in the query. Both primitives and tables are supported.
+
+    A special `$self` variable allows to perform recursion, if supported by the dialect.
+
+
     Parameters:
         result_type: The expected type of the result of the SQL query
         sql_code: The SQL code to be evaluated
@@ -148,7 +151,7 @@ def pql_fmt(state: State, s: T.string):
             assert t[0] == '$'
             obj = state.get_var(t[1:])
             inst = cast_to_instance(state, obj)
-            as_str = cast(state, inst, T.string)
+            as_str = cast(inst, T.string)
             string_parts.append(as_str)
         elif t:
             string_parts.append(objects.new_value_instance(t))
@@ -172,7 +175,8 @@ def _canonize_default(d):
 def create_internal_func(fname, f):
     sig = inspect.signature(f)
     return objects.InternalFunction(fname, [
-        objects.Param(pname, type_ if isinstance(type_, Type) else T.any, _canonize_default(sig.parameters[pname].default))
+        objects.Param(pname, type_ if isinstance(type_, Type) else T.any,
+                      _canonize_default(sig.parameters[pname].default))
         for pname, type_ in list(f.__annotations__.items())[1:]
     ], f)
 
@@ -194,6 +198,7 @@ def pql_debug(state: State):
     """Breaks the execution of the interpreter, and enters into a debug
     session using the REPL environment.
 
+    Use `c()` to continue the execution.
     """
     py_api = state._py_api
 
@@ -242,14 +247,16 @@ def _count(state, obj, table_func, name='count'):
         code = sql.FieldFunc(name, sql.AllFields(T.any))
     elif obj.type <= T.table:
         code = table_func(obj.code)
-    elif isinstance(obj, objects.StructInstance):
+    elif isinstance(obj, objects.StructInstance) and not obj.type <= T.aggregated:
+        # XXX is count() even the right method for this?
         return new_value_instance(len(obj.attrs))
 
-    elif obj.type <= T.vectorized[T.json_array]:
+    elif obj.type <= T.projected[T.json_array]:
         code = sql.JsonLength(obj.code)
     else:
-        if not (obj.type <= T.aggregate):
-            raise Signal.make(T.TypeError, None, f"Function '{name}' expected an aggregated list, but got '{obj.type}' instead. Did you forget to group?")
+        if not obj.type <= T.aggregated:
+            msg = f"Function '{name}' expected an aggregated list, but got '{obj.type}' instead. Did you forget to group?"
+            raise Signal.make(T.TypeError, None, msg)
 
         obj = obj.primary_key()
         code = sql.FieldFunc('count', obj.code)
@@ -282,7 +289,7 @@ def pql_count(state: State, obj: T.container.as_nullable() = objects.null):
     return _count(state, obj, sql.CountTable)
 
 
-def pql_temptable(state: State, expr: T.collection, const: T.bool.as_nullable() = objects.null):
+def pql_temptable(state: State, expr: T.table, const: T.bool.as_nullable() = objects.null):
     """Generate a temporary table with the contents of the given table
 
     It will remain available until the database session ends, unless manually removed.
@@ -298,7 +305,7 @@ def pql_temptable(state: State, expr: T.collection, const: T.bool.as_nullable()
     # 'temptable' creates its own counting 'id' field. Copying existing 'id' fields will cause a collision
     # 'const table' doesn't
     const = cast_to_python(state, const)
-    assert_type(expr.type, T.collection, state, expr, 'temptable')
+    assert_type(expr.type, T.table, state, expr, 'temptable')
 
     name = state.unique_name("temp")    # TODO get name from table options
 
@@ -317,7 +324,7 @@ def pql_get_db_type(state: State):
 
 
 
-def sql_bin_op(state, op, t1, t2, name, additive=False):
+def sql_bin_op(op, t1, t2, name):
 
     if not isinstance(t1, objects.CollectionInstance):
         raise Signal.make(T.TypeError, t1, f"First argument isn't a table, it's a {t1.type}")
@@ -328,37 +335,38 @@ def sql_bin_op(state, op, t1, t2, name, additive=False):
     l1 = len(t1.type.elems)
     l2 = len(t2.type.elems)
     if l1 != l2:
-        raise Signal.make(T.TypeError, None, f"Cannot {name} tables due to column mismatch (table1 has {l1} columns, table2 has {l2} columns)")
+        msg = f"Cannot {name} tables due to column mismatch (table1 has {l1} columns, table2 has {l2} columns)"
+        raise Signal.make(T.TypeError, None, msg)
 
     for e1, e2 in zip(t1.type.elems.values(), t2.type.elems.values()):
-        if not (e2 <= e1):
+        if not e2 <= e1:
             raise Signal.make(T.TypeError, None, f"Cannot {name}. Column types don't match: '{e1}' and '{e2}'")
 
     code = sql.TableArith(op, [t1.code, t2.code])
 
     return type(t1).make(code, t1.type, [t1, t2])
 
-def pql_table_intersect(state: State, t1: T.collection, t2: T.collection):
+def pql_table_intersect(_state: State, t1: T.table, t2: T.table):
     "Intersect two tables. Used for `t1 & t2`"
-    return sql_bin_op(state, "INTERSECT", t1, t2, "intersect")
+    return sql_bin_op("INTERSECT", t1, t2, "intersect")
 
-def pql_table_substract(state: State, t1: T.collection, t2: T.collection):
+def pql_table_substract(state: State, t1: T.table, t2: T.table):
     "Substract two tables (except). Used for `t1 - t2`"
     if state.db.target is sql.mysql:
         raise Signal.make(T.NotImplementedError, t1, "MySQL doesn't support EXCEPT (yeah, really!)")
-    return sql_bin_op(state, "EXCEPT", t1, t2, "subtract")
+    return sql_bin_op("EXCEPT", t1, t2, "subtract")
 
-def pql_table_union(state: State, t1: T.collection, t2: T.collection):
+def pql_table_union(_state: State, t1: T.table, t2: T.table):
     "Union two tables. Used for `t1 | t2`"
-    return sql_bin_op(state, "UNION", t1, t2, "union", True)
+    return sql_bin_op("UNION", t1, t2, "union")
 
-def pql_table_concat(state: State, t1: T.collection, t2: T.collection):
+def pql_table_concat(_state: State, t1: T.table, t2: T.table):
     "Concatenate two tables (union all). Used for `t1 + t2`"
     if isinstance(t1, objects.EmptyListInstance):
         return t2
     if isinstance(t2, objects.EmptyListInstance):
         return t1
-    return sql_bin_op(state, "UNION ALL", t1, t2, "concatenate", True)
+    return sql_bin_op("UNION ALL", t1, t2, "concatenate")
 
 
 def _get_table(t):
@@ -369,16 +377,35 @@ def _get_table(t):
         raise Signal.make(T.TypeError, None, f"join() arguments must be tables")
     return t
 
-def _join2(state, join, a, b):
+def _join2(a, b):
     if isinstance(a, objects.SelectedColumnInstance) and isinstance(b, objects.SelectedColumnInstance):
         return [a, b]
 
-    if not ((a.type <= T.collection) and (b.type <= T.collection)):
+    if not ((a.type <= T.table) and (b.type <= T.table)):
         a = a.type.repr()
         b = b.type.repr()
         raise Signal.make(T.TypeError, None, f"join() arguments must be of same type. Instead got:\n * {a}\n * {b}")
 
-    return _auto_join(state, join, a, b)
+    return _auto_join(a, b)
+
+def _auto_join(ta, tb):
+    refs1 = _find_table_reference(ta, tb)
+    refs2 = _find_table_reference(tb, ta)
+    auto_join_count = len(refs1) + len(refs2)
+    if auto_join_count < 1:
+        raise NoAutoJoinFound(ta, tb)
+    elif auto_join_count > 1:   # Ambiguity in auto join resolution
+        raise Signal.make(T.JoinError, None, "Cannot auto-join: Several plausible relations found")
+
+    if len(refs1) == 1:
+        dst, src = refs1[0]
+    elif len(refs2) == 1:
+        src, dst = refs2[0]
+    else:
+        assert False
+
+    return src, dst
+
 
 def _join(state: State, join: str, exprs_dict: dict, joinall=False, nullable=None):
 
@@ -400,7 +427,7 @@ def _join(state: State, join: str, exprs_dict: dict, joinall=False, nullable=Non
 
     # Initialization
     tables = [_get_table(x) for x in exprs]
-    assert all((t.type <= T.collection) for t in tables)
+    assert all((t.type <= T.table) for t in tables)
 
     structs = {name: T.struct(table.type.elems) for name, table in safezip(names, tables)}
 
@@ -411,10 +438,10 @@ def _join(state: State, join: str, exprs_dict: dict, joinall=False, nullable=Non
 
     tables = [objects.alias_table_columns(t, n) for n, t in safezip(names, tables)]
 
-    primary_keys = [ [name] + pk
+    primary_keys = [[name] + pk
                     for name, t in safezip(names, tables)
                     for pk in t.type.options.get('pk', [])
-                ]
+                   ]
     table_type = T.table(structs, name=Id(state.unique_name("joinall" if joinall else "join")), pk=primary_keys)
 
     conds = []
@@ -429,8 +456,9 @@ def _join(state: State, join: str, exprs_dict: dict, joinall=False, nullable=Non
         joined_exprs = set()
         for (na, ta), (nb, tb) in itertools.combinations(safezip(names, exprs), 2):
             try:
-                cols = _join2(state, join, ta, tb)
-                conds.append( sql.Compare('=', [sql.Name(c.type, join_names((n, c.name))) for n, c in safezip([na, nb], cols)]) )
+                cols = _join2(ta, tb)
+                cond = sql.Compare('=', [sql.Name(c.type, join_names((n, c.name))) for n, c in safezip([na, nb], cols)])
+                conds.append(cond)
                 joined_exprs |= {id(ta), id(tb)}
             except NoAutoJoinFound as e:
                 pass
@@ -515,34 +543,16 @@ def pql_joinall(state: State, tables):
 class NoAutoJoinFound(Exception):
     pass
 
-def _auto_join(state, join, ta, tb):
-    refs1 = _find_table_reference(ta, tb)
-    refs2 = _find_table_reference(tb, ta)
-    auto_join_count = len(refs1) + len(refs2)
-    if auto_join_count < 1:
-        raise NoAutoJoinFound(ta, tb)
-    elif auto_join_count > 1:   # Ambiguity in auto join resolution
-        raise Signal.make(T.JoinError, None, "Cannot auto-join: Several plausible relations found")
-
-    if len(refs1) == 1:
-        dst, src = refs1[0]
-    elif len(refs2) == 1:
-        src, dst = refs2[0]
-    else:
-        assert False
-
-    return src, dst
-
 
 @listgen
 def _find_table_reference(t1, t2):
     for name, c in t1.type.elems.items():
-        if (c <= T.t_relation):
+        if c <= T.t_relation:
             rel = c.options['rel']
             if rel['table'] == t2.type:     # if same table
                 yield t2.get_attr(rel['column']), objects.SelectedColumnInstance(t1, c, name)
 
-def pql_type(state: State, obj: T.any):
+def pql_type(_state: State, obj: T.any):
     """Returns the type of the given object
 
     Example:
@@ -559,29 +569,31 @@ def pql_type(state: State, obj: T.any):
 def pql_repr(state: State, obj: T.any):
     """Returns the representation text of the given object
     """
+    if obj.type <= T.projected | T.aggregated:
+        raise Signal.make(T.CompileError, obj, "repr() cannot run in projected/aggregated mode")
+
     try:
         return new_value_instance(obj.repr())
     except ValueError:
         value = repr(cast_to_python(state, obj))
         return new_value_instance(value)
 
-def pql_columns(state: State, table: T.collection):
+def pql_columns(_state: State, obj: T.container):
     """Returns a dictionary `{column_name: column_type}` for the given table
 
     Example:
         >> columns([0])
         {item: int}
     """
-    # TODO table: T.table, not collection
 
-    elems = table.type.elems
+    elems = obj.type.elems
     if isinstance(elems, tuple):    # Create a tuple/list instead of dict?
         elems = {f't{i}':e for i, e in enumerate(elems)}
 
     return ast.Dict_(elems)
 
 
-def pql_cast(state: State, obj: T.any, target_type: T.type):
+def pql_cast(_state: State, obj: T.any, target_type: T.type):
     """Attempt to cast an object to a specified type
 
     The resulting object will be of type `target_type`, or a `TypeError`
@@ -599,7 +611,7 @@ def pql_cast(state: State, obj: T.any, target_type: T.type):
     if obj.type is type_:
         return obj
 
-    return cast(state, obj, type_)
+    return cast(obj, type_)
 
 
 def pql_import_table(state: State, name: T.string, columns: T.list[T.string].as_nullable() = objects.null):
@@ -647,20 +659,19 @@ def pql_connect(state: State, uri: T.string, load_all_tables: T.bool = ast.false
         state._py_api.load_all_tables()     # XXX
     return objects.null
 
-def pql_help(state: State, inst: T.any = objects.null):
+def pql_help(_state: State, inst: T.any = objects.null):
     """Provides a brief summary for the given object
     """
     if inst is objects.null:
         text = (
-            "Welcome to Preql!\n\n"
-            "To see the list of functions and objects available in the namespace, type 'names()'\n"
+            "To see the list of functions and objects available in the namespace, type '[b]names()[/b]'\n"
             "To see the next page of a table preview, type '.' and then enter\n"
             "\n"
-            "To get help for a specific function, type 'help(func_object)'\n\n"
+            "To get help for a specific function, type '[b]help(an_object)[/b]'\n\n"
             "For example:\n"
             "    >> help(help)\n"
         )
-        return new_value_instance(text, T.string).replace(type=T.text)
+        return new_value_instance(text, T.string).replace(type=T._rich)
 
 
     lines = []
@@ -672,6 +683,8 @@ def pql_help(state: State, inst: T.any = objects.null):
         lines = [f"No help available for {inst.repr()}"]
     except runtype.DispatchError:
         lines += [f"<doc not available yet for object of type '{inst.type}>'"]
+    except AutoDocError:
+        lines += [f"<error generating documentation for object '{inst}'"]
 
 
     text = '\n'.join(lines) + '\n'
@@ -809,12 +822,10 @@ def pql_import_csv(state: State, table: T.table, filename: T.string, header: T.b
         header: If true, skips the first line
     """
     # TODO better error handling, validation
-    from tqdm import tqdm   # TODO use rich instead?
-
 
     filename = cast_to_python(state, filename)
     header = cast_to_python(state, header)
-    print(f"Importing CSV file: '{filename}'")
+    msg = f"Importing CSV file: '{filename}'"
 
     ROWS_PER_QUERY = 1024
 
@@ -827,31 +838,34 @@ def pql_import_csv(state: State, table: T.table, filename: T.string, header: T.b
         db_query(state, q)
 
 
-    with open(filename, 'r', encoding='utf8') as f:
-        line_count = len(list(f))
-        f.seek(0)
+    try:
+        with open(filename, 'r', encoding='utf8') as f:
+            line_count = len(list(f))
+            f.seek(0)
 
-        reader = csv.reader(f)
-        for i, row in enumerate(tqdm(reader, total=line_count)):
-            if i == 0:
-                matched = cons.match_params(state, row)
-                keys = [p.name for (p, _) in matched]
+            reader = csv.reader(f)
+            for i, row in enumerate(rich.progress.track(reader, total=line_count, description=msg)):
+                if i == 0:
+                    matched = cons.match_params(state, row)
+                    keys = [p.name for (p, _) in matched]
 
-            if header and i == 0:
-                # Skip first line if header=True
-                continue
+                if header and i == 0:
+                    # Skip first line if header=True
+                    continue
 
-            values = ["'%s'" % (v.replace("'", "''")) for v in row]
-            rows.append(values)
+                values = ["'%s'" % (v.replace("'", "''")) for v in row]
+                rows.append(values)
 
-            if (i+1) % ROWS_PER_QUERY == 0:
-                insert_values()
-                rows = []
+                if (i+1) % ROWS_PER_QUERY == 0:
+                    insert_values()
+                    rows = []
 
-    if keys and rows:
-        insert_values()
+        if keys and rows:
+            insert_values()
 
-    return table
+        return table
+    except FileNotFoundError as e:
+        raise Signal.make(T.FileError, None, str(e))
 
 
 def _rest_func_endpoint(state, func):
@@ -912,7 +926,7 @@ def pql_serve_rest(state: State, endpoints: T.struct, port: T.int = new_int(8080
 
     port_ = cast_to_python(state, port)
 
-    async def root(request):
+    async def root(_request):
         return JSONResponse(list(endpoints.attrs))
 
     routes = [
@@ -926,7 +940,7 @@ def pql_serve_rest(state: State, endpoints: T.struct, port: T.int = new_int(8080
                 path += "/{%s}" % p.name
 
             routes.append(Route(path, endpoint=_rest_func_endpoint(state, func)))
-        elif func.type <= T.collection:
+        elif func.type <= T.table:
             routes.append(Route(path, endpoint=_rest_table_endpoint(state, func)))
         else:
             raise Signal.make(T.TypeError, func, f"Expected a function or a table, got {func.type}")
